/* auditd-audispd_conf.c -- 
 * Copyright 2004-2006 Red Hat Inc., Durham, North Carolina.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *   Steve Grubb <sgrubb@redhat.com>
 * 
 */
#define EX_UTILS_NO_FUNCS 1
#include "ex_utils.h"

#include "mk.h"

#include "auditd_audispd_conf.h"

/* NOTE: mostly borrowed from auditd-config.c so the config. files are the
   same */

/* def. flags for parsing numbers... */
#define NFLAGS (10 | VSTR_FLAG_PARSE_NUM_SEP | VSTR_FLAG_PARSE_NUM_NO_BEG_PM)

struct nv_pair
{
 size_t npos;
 size_t nlen;
 
 size_t vpos;
 size_t vlen;
};

struct kwm_pair
{
 const char *name;
 int (*parser)(const Vstr_base *, size_t, size_t, int, Audispd_opts *);
};

struct kwc_pair
{
 const char *name;
 int (*parser)(const Vstr_base *, size_t, size_t, int, Audispd_cmd *);
};

struct kwp_pair
{
 const char *name;
 int (*parser)(const Vstr_base *, size_t, size_t, int, Audispd_policy_opts *);
};

static int auditd_audispd__main_cntl_file(const Vstr_base *val,
                                          size_t pos, size_t len,
                                          int COMPILE_ATTR_UNUSED(line),
                                          Audispd_opts *opts)
{
  Vstr_base *s1 = NULL;

  s1 = opts->s->cntl_file;
  
  return !vstr_sub_vstr(s1, 1, s1->len, val, pos, len, VSTR_TYPE_SUB_DEF);
}

static int auditd_audispd__main_main_input(const Vstr_base *val,
                                           size_t pos, size_t len,
                                           int COMPILE_ATTR_UNUSED(line),
                                           Audispd_opts *opts)
{
  Vstr_base *s1 = NULL;

  s1 = opts->main_input;
  
  return !vstr_sub_vstr(s1, 1, s1->len, val, pos, len, VSTR_TYPE_SUB_DEF);
}

static int auditd_audispd__main_daemonize(const Vstr_base *val,
                                          size_t pos, size_t len,
                                          int COMPILE_ATTR_UNUSED(line),
                                          Audispd_opts *opts)
{
  if (FALSE) { }

  else if (vstr_cmp_case_cstr_eq(val, pos, len, "true"))
    opts->s->become_daemon = TRUE;
  else if (vstr_cmp_case_cstr_eq(val, pos, len, "false"))
    opts->s->become_daemon = FALSE;
  else if (vstr_cmp_case_cstr_eq(val, pos, len, "on"))
    opts->s->become_daemon = TRUE;
  else if (vstr_cmp_case_cstr_eq(val, pos, len, "off"))
    opts->s->become_daemon = FALSE;

  else return 1;
    
  return 0;
}

static int auditd_audispd__main_unman_con(const Vstr_base *val,
                                          size_t pos, size_t len,
                                          int COMPILE_ATTR_UNUSED(line),
                                          Audispd_opts *opts)
{
  Vstr_base *s1 = NULL;
  
  s1 = opts->s->addr_beg->acpt_address;
  
  return !vstr_sub_vstr(s1, 1, s1->len, val, pos, len, VSTR_TYPE_SUB_DEF);
}

static int auditd_audispd__main_unman_pol(const Vstr_base *val,
                                          size_t pos, size_t len,
                                          int COMPILE_ATTR_UNUSED(line),
                                          Audispd_opts *opts)
{
  Vstr_base *s1 = NULL;
  
  s1 = opts->s->addr_beg->def_policy;
  
  return !vstr_sub_vstr(s1, 1, s1->len, val, pos, len, VSTR_TYPE_SUB_DEF);
}

static int auditd_audispd__main_unman_qlen(const Vstr_base *val,
                                           size_t pos, size_t len,
                                           int COMPILE_ATTR_UNUSED(line),
                                           Audispd_opts *opts)
{
  if (vstr_cmp_case_cstr_eq(val, pos, len, "<none>"))
    return (opts->s->addr_beg->q_listen_len = 0);

  opts->s->addr_beg->q_listen_len = vstr_parse_uint(val, pos, len, NFLAGS,
                                                    NULL, NULL);
  return (!opts->s->addr_beg->q_listen_len);
}

static int auditd_audispd__main_unman_cmax(const Vstr_base *val,
                                           size_t pos, size_t len,
                                           int COMPILE_ATTR_UNUSED(line),
                                           Audispd_opts *opts)
{
  if (vstr_cmp_case_cstr_eq(val, pos, len, "<none>"))
    return (opts->s->addr_beg->max_connections = 0);
  if (vstr_cmp_case_cstr_eq(val, pos, len, "<unlimited>"))
    return (opts->s->addr_beg->max_connections = 0);

  opts->s->addr_beg->max_connections = vstr_parse_uint(val, pos, len, NFLAGS,
                                                       NULL, NULL);
  return (!opts->s->addr_beg->max_connections);
}

static const struct kwm_pair main_keywords[] = 
{
  {"cntl-file", auditd_audispd__main_cntl_file },
  {"daemonize", auditd_audispd__main_daemonize },
  {"main-input", auditd_audispd__main_main_input },
  
  {"unmanaged-connections-path",         auditd_audispd__main_unman_con },
  {"unmanaged-connections-policy",       auditd_audispd__main_unman_pol },
  {"unmanaged-connections-queue-length", auditd_audispd__main_unman_qlen },
  {"unmanaged-connections-max",          auditd_audispd__main_unman_cmax },
  
  { NULL, NULL }
};


static int auditd_audispd__cmd_path(const Vstr_base *val,
                                    size_t pos, size_t len,
                                    int COMPILE_ATTR_UNUSED(line),
                                    Audispd_cmd *cmd)
{
  Vstr_base *s1 = NULL;

  s1 = cmd->path;
  
  return !vstr_sub_vstr(s1, 1, s1->len, val, pos, len, VSTR_TYPE_SUB_DEF);
}

static int auditd_audispd__cmd_arg1(const Vstr_base *val,
                                    size_t pos, size_t len,
                                    int COMPILE_ATTR_UNUSED(line),
                                    Audispd_cmd *cmd)
{
  Vstr_base *s1 = NULL;

  s1 = cmd->args[0];
  
  cmd->arg_num = 1;  
  return !vstr_sub_vstr(s1, 1, s1->len, val, pos, len, VSTR_TYPE_SUB_DEF);
}

static int auditd_audispd__cmd_arg2(const Vstr_base *val,
                                    size_t pos, size_t len,
                                    int COMPILE_ATTR_UNUSED(line),
                                    Audispd_cmd *cmd)
{
  Vstr_base *s1 = NULL;

  s1 = cmd->args[1];
  
  cmd->arg_num = 2;
  return !vstr_sub_vstr(s1, 1, s1->len, val, pos, len, VSTR_TYPE_SUB_DEF);
}

static int auditd_audispd__cmd_arg3(const Vstr_base *val,
                                    size_t pos, size_t len,
                                    int COMPILE_ATTR_UNUSED(line),
                                    Audispd_cmd *cmd)
{
  Vstr_base *s1 = NULL;

  s1 = cmd->args[2];
  
  cmd->arg_num = 3;
  return !vstr_sub_vstr(s1, 1, s1->len, val, pos, len, VSTR_TYPE_SUB_DEF);
}

static int auditd_audispd__cmd_arg4(const Vstr_base *val,
                                    size_t pos, size_t len,
                                    int COMPILE_ATTR_UNUSED(line),
                                    Audispd_cmd *cmd)
{
  Vstr_base *s1 = NULL;

  s1 = cmd->args[3];
  
  cmd->arg_num = 4;
  return !vstr_sub_vstr(s1, 1, s1->len, val, pos, len, VSTR_TYPE_SUB_DEF);
}

static int auditd_audispd__cmd_arg5(const Vstr_base *val,
                                    size_t pos, size_t len,
                                    int COMPILE_ATTR_UNUSED(line),
                                    Audispd_cmd *cmd)
{
  Vstr_base *s1 = NULL;

  s1 = cmd->args[4];
  
  cmd->arg_num = 5;
  return !vstr_sub_vstr(s1, 1, s1->len, val, pos, len, VSTR_TYPE_SUB_DEF);
}

static int auditd_audispd__cmd_arg6(const Vstr_base *val,
                                    size_t pos, size_t len,
                                    int COMPILE_ATTR_UNUSED(line),
                                    Audispd_cmd *cmd)
{
  Vstr_base *s1 = NULL;

  s1 = cmd->args[5];
  
  cmd->arg_num = 6;
  return !vstr_sub_vstr(s1, 1, s1->len, val, pos, len, VSTR_TYPE_SUB_DEF);
}

static int auditd_audispd__cmd_arg7(const Vstr_base *val,
                                    size_t pos, size_t len,
                                    int COMPILE_ATTR_UNUSED(line),
                                    Audispd_cmd *cmd)
{
  Vstr_base *s1 = NULL;

  s1 = cmd->args[6];
  
  cmd->arg_num = 7;
  return !vstr_sub_vstr(s1, 1, s1->len, val, pos, len, VSTR_TYPE_SUB_DEF);
}

static int auditd_audispd__cmd_arg8(const Vstr_base *val,
                                    size_t pos, size_t len,
                                    int COMPILE_ATTR_UNUSED(line),
                                    Audispd_cmd *cmd)
{
  Vstr_base *s1 = NULL;

  s1 = cmd->args[7];
  
  cmd->arg_num = 8;
  return !vstr_sub_vstr(s1, 1, s1->len, val, pos, len, VSTR_TYPE_SUB_DEF);
}

static int auditd_audispd__cmd_policy(const Vstr_base *val,
                                      size_t pos, size_t len,
                                      int COMPILE_ATTR_UNUSED(line),
                                      Audispd_cmd *cmd)
{
  Vstr_base *s1 = NULL;

  s1 = cmd->def_policy;
  
  return !vstr_sub_vstr(s1, 1, s1->len, val, pos, len, VSTR_TYPE_SUB_DEF);
}

static const struct kwc_pair cmd_keywords[] = 
{
  {"path",  auditd_audispd__cmd_path },
  {"arg1",  auditd_audispd__cmd_arg1 },
  {"arg2",  auditd_audispd__cmd_arg2 },
  {"arg3",  auditd_audispd__cmd_arg3 },
  {"arg4",  auditd_audispd__cmd_arg4 },
  {"arg5",  auditd_audispd__cmd_arg5 },
  {"arg6",  auditd_audispd__cmd_arg6 },
  {"arg7",  auditd_audispd__cmd_arg7 },
  {"arg8",  auditd_audispd__cmd_arg8 },
  {"policy", auditd_audispd__cmd_policy },
  { NULL, NULL }
};

static int auditd_audispd__pol_lim_data(const Vstr_base *val,
                                        size_t pos, size_t len,
                                        int COMPILE_ATTR_UNUSED(line),
                                        Audispd_policy_opts *opts)
{
  if (vstr_cmp_case_cstr_eq(val, pos, len, "<none>"))
    return (opts->max_data = 0);
  if (vstr_cmp_case_cstr_eq(val, pos, len, "<unlimited>"))
    return (opts->max_data = 0);
  
  opts->max_data = vstr_parse_uint(val, pos, len, NFLAGS, NULL, NULL);
  
  return !opts->max_data;
}

static int auditd_audispd__pol_lim_reach(const Vstr_base *val,
                                         size_t pos, size_t len,
                                         int COMPILE_ATTR_UNUSED(line),
                                         Audispd_policy_opts *opts)
{
  if (FALSE) { }

  else if (vstr_cmp_case_cstr_eq(val, pos, len, "drop-messages"))
    opts->limit_reached = AUDISPD_TYPE_LIM_REACHED_DROP;
  else if (vstr_cmp_case_cstr_eq(val, pos, len, "close-connection"))
    opts->limit_reached = AUDISPD_TYPE_LIM_REACHED_CLOSE;
  else if (FALSE && vstr_cmp_case_cstr_eq(val, pos, len, "stop-input"))
    opts->limit_reached = AUDISPD_TYPE_LIM_REACHED_STOP;

  else return 1;
  
  return 0;
}

static int auditd_audispd__pol_man_con_ded(const Vstr_base *val,
                                           size_t pos, size_t len,
                                           int COMPILE_ATTR_UNUSED(line),
                                           Audispd_policy_opts *opts)
{
  if (FALSE) { }

  else if (vstr_cmp_case_cstr_eq(val, pos, len, "restart"))
    opts->managed_con_died = AUDISPD_TYPE_MAN_CON_DIED_RESTART;
  else if (vstr_cmp_case_cstr_eq(val, pos, len, "ignore"))
    opts->managed_con_died = AUDISPD_TYPE_MAN_CON_DIED_IGNORE;
  else if (vstr_cmp_case_cstr_eq(val, pos, len, "cascade"))
    opts->managed_con_died = AUDISPD_TYPE_MAN_CON_DIED_CASCADE;

  else return 1;
    
  return 0;
}

static const struct kwp_pair pol_keywords[] = 
{
  {"limit-data",  auditd_audispd__pol_lim_data },
  {"limit-reached",   auditd_audispd__pol_lim_reach },
  {"managed-connection-death", auditd_audispd__pol_man_con_ded },
  { NULL, NULL }
};

static int auditd_audispd__read_file_check(Vstr_base *fdata, Audispd_opts *opts,
                                           Vstr_base *out, const char *fname,
                                           int must_be_uid_zero)
{
  int fd = -1;
  struct stat st;

  vstr_del(fdata, 1, fdata->len);
  
  fd = open(fname, O_RDONLY | O_NOCTTY);
  if (fd < 0)
  {
    vstr_add_fmt(out, out->len, "open(%s): %m", fname);
    goto open_file_fail;
  }
  
  /* check the file's permissions: owned by root, not world writable,
   * not symlink.
   */
  if (fstat(fd, &st) < 0)
  {
    vstr_add_fmt(out, out->len, "fstat(%s): %m", fname);
    goto check_file_fail;
  }
  if (must_be_uid_zero && (st.st_uid != 0))
  {
    vstr_add_fmt(out, out->len, "fstat(%s)->st_uid != 0", fname);
    goto check_file_fail;
  }
  if ((st.st_mode & S_IWOTH) == S_IWOTH)
  {
    vstr_add_fmt(out, out->len, "%s is world writable", fname);
    goto check_file_fail;
  }
  if (st.st_size > (100 * 1024)) /* 100KB max file size */
  {
    vstr_add_fmt(out, out->len, "%s is too big", fname);
    goto check_file_fail;
  }  
  ASSERT(st.st_size >= 0);
  
  while (fdata->len < (size_t)st.st_size) /* read entire file... */
  {
    if (!vstr_sc_read_len_fd(fdata, fdata->len, fd,
                             st.st_size - fdata->len, NULL))
    {
      vstr_add_fmt(out, out->len, "fdopen(%s): %m", fname);
      goto read_file_fail;
    }
  }
  close(fd);
  
  if (vstr_cmp_bod_cstr_eq(fdata, 1, fdata->len, "#! /usr/bin/and-conf"))
  {
    if (!audispd_conf_parse_file(out, opts, fname))
      return (FALSE);
    
    vstr_del(fdata, 1, fdata->len);
    vstr_add_cstr_buf(fdata, 0, "\n"); /* kind of hacky, but... */
  }

  return (TRUE);

 read_file_fail:
 check_file_fail:
  close(fd);
 open_file_fail:
  return (FALSE);
}

static int get_line(Vstr_base *s1, size_t *pos, size_t *len, int *lineno)
{
  /* assume only *pos is valid... work out *len. */
  *len = vstr_sc_posdiff(*pos, s1->len);

  while (TRUE)
  {
    /* skip lines... */
    size_t tmp = vstr_spn_cstr_chrs_fwd(s1, *pos, *len, "\t ");
    *len -= tmp; *pos += tmp;

    if (tmp)
      continue;
    
    tmp = vstr_spn_cstr_chrs_fwd(s1, *pos, *len, "\n");
    *len -= tmp; *pos += tmp;
    *lineno += tmp;

    if (!tmp)
      break;
  }
  
  /* find length of current line... */
  *len = vstr_cspn_cstr_chrs_fwd(s1, *pos, *len, "\n");

  return (!!*len);
}

static int nv_split(Vstr_base *s1, size_t pos, size_t len, struct nv_pair *nv)
{
  size_t tmp = 0;

  ASSERT(!vstr_spn_cstr_chrs_fwd(s1, pos, len, "\t "));

  if (vstr_export_chr(s1, pos) == '#')
    return -1; /* If there's a comment, go to next line */

  tmp = vstr_cspn_cstr_chrs_fwd(s1, pos, len, "\t ");
  
  nv->npos = pos;
  nv->nlen = tmp;

  /* skip WS */
  len -= tmp; pos += tmp;
  tmp = vstr_spn_cstr_chrs_fwd(s1, pos, len, "\t ");
  len -= tmp; pos += tmp;

  /* get the name */
  tmp = vstr_cspn_cstr_chrs_fwd(s1, pos, len, "\t ");
  if (!tmp)
    return 1;
  if (!vstr_cmp_cstr_eq(s1, pos, tmp, "="))
    return 2;

  /* skip WS */
  len -= tmp; pos += tmp;
  tmp = vstr_spn_cstr_chrs_fwd(s1, pos, len, "\t ");
  len -= tmp; pos += tmp;

  /* get the value */
  if (!len)
    return 1;

  /* delete WS at end... */
  len -= vstr_spn_cstr_chrs_rev(s1, pos, len, "\t\r ");

  nv->vpos = pos;
  nv->vlen = len;

  /* Everything is OK */
  return 0;
}

static void auditd_audispd__conf_rc(int rc, Vstr_base *out,
                                    int lineno, const char *fname)
{
  switch (rc)
  {
    case -1: /* skip */
    case 0: /* fine */
      break;
    case 1: /* not the right number of tokens. */
      vstr_add_fmt(out, out->len,
                   "Wrong number of arguments for line %d in %s", 
                   lineno, fname);
      break;
    case 2: /* no '=' sign */
      vstr_add_fmt(out, out->len, "Missing equal sign for line %d in %s", 
                   lineno, fname);
      break;
    default: /* something else went wrong...  */
      vstr_add_fmt(out, out->len, "Unknown error for line %d in %s", 
                   lineno, fname);
      break;
  }
}

static const struct kwm_pair *kwm_lookup(const Vstr_base *s1,
                                         size_t pos, size_t len)
{
        unsigned int scan = 0;
        
        while (main_keywords[scan].name != NULL) {
                if (vstr_cmp_case_cstr_eq(s1, pos, len,
                                          main_keywords[scan].name))
                        break;
                ++scan;
        }
        return &main_keywords[scan];
}

static const struct kwc_pair *kwc_lookup(const Vstr_base *s1,
                                         size_t pos, size_t len)
{
        unsigned int scan = 0;
        
        while (cmd_keywords[scan].name != NULL) {
                if (vstr_cmp_case_cstr_eq(s1, pos, len,
                                          cmd_keywords[scan].name))
                        break;
                ++scan;
        }
        return &cmd_keywords[scan];
}

static const struct kwp_pair *kwp_lookup(const Vstr_base *s1,
                                         size_t pos, size_t len)
{
        unsigned int scan = 0;
        
        while (pol_keywords[scan].name != NULL) {
                if (vstr_cmp_case_cstr_eq(s1, pos, len,
                                          pol_keywords[scan].name))
                        break;
                ++scan;
        }
        return &pol_keywords[scan];
}

static int auditd_audispd__fail_bad_key(Vstr_base *out, Vstr_base *fdata,
                                        const struct nv_pair *nv,
                                        int lineno, const char *fname)
{
  vstr_add_fmt(out, out->len, "Unknown keyword \"${vstr.buf:%p%zu%zu%u}\""
               " at line %d of file %s",
               fdata, nv->npos, nv->nlen, VSTR_TYPE_ADD_DEF, lineno, fname);
  vstr_free_base(fdata);
  return (FALSE);
}

static int auditd_audispd__fail_bad_arg(Vstr_base *out, Vstr_base *fdata,
                                        const struct nv_pair *nv,
                                        int lineno, const char *fname)
{
  vstr_add_fmt(out, out->len, "Bad value \"${vstr.buf:%p%zu%zu%u}\""
               " for keyword \"${vstr.buf:%p%zu%zu%u}\""
               " at line %d of file %s",
               fdata, nv->npos, nv->nlen, VSTR_TYPE_ADD_DEF,
               fdata, nv->vpos, nv->vlen, VSTR_TYPE_ADD_DEF,
               lineno, fname);
  vstr_free_base(fdata);
  return (FALSE);
}

int auditd_audispd_main_conf_parse_file(Vstr_base *out, Audispd_opts *opts,
                                        const char *fname)
{
  Vstr_base *fdata = NULL;
  int lineno = 1;
  size_t pos = 0;
  size_t len = 0;
  
  if (!(fdata = vstr_make_base(out->conf)))
    goto conf_file_init_fail;

  if (!auditd_audispd__read_file_check(fdata, opts, out, fname, TRUE))
    goto read_file_fail;

  pos = 1;
  while (get_line(fdata, &pos, &len, &lineno))
  {
    const struct kwm_pair *kw;
    struct nv_pair nv;
    int rc = -1;

    rc = nv_split(fdata, pos, len, &nv);
    auditd_audispd__conf_rc(rc, out, lineno, fname);

    if (rc == -1)
    {
      pos += len;
      continue;
    }
    
    if (rc != 0)
      goto parse_file_fail;

    /* identify keyword or error */
    kw = kwm_lookup(fdata, nv.npos, nv.nlen);
    if (kw->name == NULL)
      return (auditd_audispd__fail_bad_key(out, fdata, &nv, lineno, fname));
    
    /* dispatch to keyword's local parser */
    if (kw->parser(fdata, nv.vpos, nv.vlen, lineno, opts))
      return (auditd_audispd__fail_bad_arg(out, fdata, &nv, lineno, fname));

    pos += len;
  }
  vstr_free_base(fdata);
  
  return (TRUE);

 read_file_fail:
 parse_file_fail:
  vstr_free_base(fdata);
  return (FALSE);
  
 conf_file_init_fail:
  vstr_add_fmt(out, out->len, " Couln't init config. file %s", fname);    
  return (FALSE);
}

int auditd_audispd_cmd_conf_parse_file(Vstr_base *out, Audispd_opts *opts,
                                       const char *fname)
{
  Vstr_base *fdata = NULL;
  int lineno = 1;
  size_t pos = 0;
  size_t len = 0;
  unsigned int scan = 0;
  Audispd_cmd *cmd = NULL;
  const char *bname = NULL;

  if (!(fdata = vstr_make_base(out->conf)))
    goto conf_file_init_fail;

  if (!auditd_audispd__read_file_check(fdata, opts, out, fname, FALSE))
    goto read_file_fail;

  if (vstr_cmp_cstr_eq(fdata, 1, fdata->len, "\n"))
    goto skip_new_cmd; /* did it via. audispd_conf.c */
  
  if ((bname = strrchr(fname, '/')))
    ++bname;
  else
    bname = fname;
  
  while (scan < opts->cmd_num)
  {
    cmd = &opts->cmds[scan++];
    if (vstr_cmp_cstr_eq(cmd->name, 1, cmd->name->len, bname))
      break;
    cmd = NULL;
  }
  if (!cmd && (opts->cmd_num < AUDISPD_CONF_MAX_CMDS))
  {
    cmd = &opts->cmds[opts->cmd_num++];
    
    ASSERT(!cmd->name->len);
    if (!vstr_add_cstr_buf(cmd->name, 0, bname))
      return (FALSE);
  }
  if (!cmd)
    return (FALSE);
  
  pos = 1;
  while (get_line(fdata, &pos, &len, &lineno))
  {
    const struct kwc_pair *kw;
    struct nv_pair nv;
    int rc = -1;

    rc = nv_split(fdata, pos, len, &nv);
    auditd_audispd__conf_rc(rc, out, lineno, fname);

    if (rc == -1)
    {
      pos += len;
      continue;
    }
    
    if (rc != 0)
      goto parse_file_fail;

    /* identify keyword or error */
    kw = kwc_lookup(fdata, nv.npos, nv.nlen);
    if (kw->name == NULL)
      return (auditd_audispd__fail_bad_key(out, fdata, &nv, lineno, fname));
    
    /* dispatch to keyword's local parser */
    if (kw->parser(fdata, nv.vpos, nv.vlen, lineno, cmd))
      return (auditd_audispd__fail_bad_arg(out, fdata, &nv, lineno, fname));

    pos += len;
  }

 skip_new_cmd:
  vstr_free_base(fdata);
  
  return (TRUE);
  
 read_file_fail:
 parse_file_fail:
  vstr_free_base(fdata);
  return (FALSE);
  
 conf_file_init_fail:
  vstr_add_fmt(out, out->len, " Couln't init config. file %s", fname);    
  return (FALSE);
}

int auditd_audispd_sc_cmd_conf_parse_dir_file(Vstr_base *out,
                                              void *opts,
                                              const char *fname)
{
  return (auditd_audispd_cmd_conf_parse_file(out, opts, fname));
}

static int auditd_audispd_pol_conf_parse_file(Vstr_base *out,
                                              Audispd_opts *opts,
                                              const char *fname)
{
  Vstr_base *fdata = NULL;
  int lineno = 1;
  size_t pos = 0;
  size_t len = 0;
  Opt_serv_policy_opts *popts = NULL;
  
  if (!(fdata = vstr_make_base(out->conf)))
    goto conf_file_init_fail;

  if (!vstr_add_cstr_ptr(fdata, 0, fname))
    goto conf_file_add_fail;

  vstr_sc_basename(fdata, 1, fdata->len, &pos, &len);

  if (!(popts = opt_policy_find(opts->s, fdata, pos, len)))
  {
    Vstr_base *name = NULL;
    
    if (!(popts = opts->s->make_policy(opts->s)))
      goto conf_file_add_fail;
    if (!(*opts->s->copy_policy)(popts, opts->s->def_policy))
      goto conf_file_add_fail;
    
    name = popts->policy_name;
    if (!vstr_sub_vstr(name, 1, name->len, fdata, pos, len, VSTR_TYPE_SUB_DEF))
      goto conf_file_add_fail;
  }
    
  if (!auditd_audispd__read_file_check(fdata, opts, out, fname, FALSE))
    goto read_file_fail;

  pos = 1;
  while (get_line(fdata, &pos, &len, &lineno))
  {
    const struct kwp_pair *kw;
    struct nv_pair nv;
    int rc = -1;

    rc = nv_split(fdata, pos, len, &nv);
    auditd_audispd__conf_rc(rc, out, lineno, fname);

    if (rc == -1)
    {
      pos += len;
      continue;
    }
    
    if (rc != 0)
      goto parse_file_fail;

    /* identify keyword or error */
    kw = kwp_lookup(fdata, nv.npos, nv.nlen);
    if (kw->name == NULL)
      return (auditd_audispd__fail_bad_key(out, fdata, &nv, lineno, fname));
    
    /* dispatch to keyword's local parser */
    if (kw->parser(fdata, nv.vpos, nv.vlen, lineno,
                   (Audispd_policy_opts *)popts))
      return (auditd_audispd__fail_bad_arg(out, fdata, &nv, lineno, fname));

    pos += len;
  }
  vstr_free_base(fdata);
  
  return (TRUE);

 read_file_fail:
 parse_file_fail:
  vstr_free_base(fdata);
  return (FALSE);
  
 conf_file_add_fail:
  vstr_free_base(fdata);
 conf_file_init_fail:
  vstr_add_fmt(out, out->len, " Couln't init config. file %s", fname);    
  return (FALSE);
}

int auditd_audispd_sc_pol_conf_parse_dir_file(Vstr_base *out,
                                              void *opts,
                                              const char *fname)
{
  return (auditd_audispd_pol_conf_parse_file(out, opts, fname));
}
