/*
 *  Copyright (C) 2004, 2005, 2006, 2007  James Antill
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  email: james@and.org
 */
/* conditionally compliant HTTP/1.1 server. */
#define _GNU_SOURCE 1 /* strsignal() / posix_fadvise64 */
#include <vstr.h>

#include <socket_poll.h>
#include <timer_q.h>

#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <err.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include <grp.h>

#define EX_UTILS_NO_USE_INIT  1
#define EX_UTILS_NO_USE_EXIT  1
#define EX_UTILS_NO_USE_LIMIT 1
#define EX_UTILS_NO_USE_OPEN  1
#define EX_UTILS_NO_USE_GET   1
#define EX_UTILS_NO_USE_IO_FD 1
#define EX_UTILS_RET_FAIL     1
#include "ex_utils.h"

#include "mk.h"

MALLOC_CHECK_DECL();

#include "cntl.h"
#include "date.h"

#define AUDISP2FILE_HAVE_GLOBAL_OPTS 1
#include "plugin-audisp2file.h"

#include <syslog.h>

AUDISP2FILE_CONF_DECL_OPTS(audisp2file_opts);

#ifndef TCP_CONGESTION
# ifdef __linux__
#  define TCP_CONGESTION 13
# else
#  define TCP_CONGESTION 0
# endif
#endif

/* we get prctl.h from evnt.h */
#if defined(HAVE_SYS_CAPABILITY_H) && defined(PR_SET_KEEPCAPS)
# include <sys/capability.h>
# define PROC_CNTL_KEEPCAPS(x1) prctl(PR_SET_KEEPCAPS, x1, 0, 0, 0)
#else
# define PROC_CNTL_KEEPCAPS(x1) (errno = ENOSYS, -1)
# define cap_t void *
# define cap_from_text(x) (errno = ENOSYS, NULL)
# define cap_set_proc(x) (errno = ENOSYS, -1)
# define cap_free(x) (errno = ENOSYS, -1)
#endif

#ifdef PR_SET_DUMPABLE
# define PROC_CNTL_DUMPABLE(x1) prctl(PR_SET_DUMPABLE, x1, 0, 0, 0)
#else
# define PROC_CNTL_DUMPABLE(x1) (errno = ENOSYS, -1)
#endif




#define CLEN COMPILE_STRLEN

/* is the cstr a prefix of the vstr */
#define VPREFIX(vstr, p, l, cstr)                                       \
    (((l) >= CLEN(cstr)) &&                                             \
     vstr_cmp_buf_eq(vstr, p, CLEN(cstr), cstr, CLEN(cstr)))
/* is the cstr a suffix of the vstr */
#define VSUFFIX(vstr, p, l, cstr)                                       \
    (((l) >= CLEN(cstr)) &&                                             \
     vstr_cmp_eod_buf_eq(vstr, p, l, cstr, CLEN(cstr)))

/* is the cstr a prefix of the vstr, no case */
#define VIPREFIX(vstr, p, l, cstr)                                      \
    (((l) >= CLEN(cstr)) &&                                             \
     vstr_cmp_case_buf_eq(vstr, p, CLEN(cstr), cstr, CLEN(cstr)))

/* for simplicity */
#define VALLEQ(vstr, cstr) vstr_cmp_cstr_eq(vstr, 1, (vstr)->len, cstr)


static Vlg *vlg = NULL;

static void usage(const char *program_name, int ret, const char *prefix)
{
  Vstr_base *out = vstr_make_base(NULL);

  if (!out)
    errno = ENOMEM, err(EXIT_FAILURE, "usage");

  vstr_add_fmt(out, 0, "%s\n\
 Format: %s [options] [file]\n\
  Daemon options\n\
    --configuration-file\n\
    --config-file  -C - Load configuration file specified.\n\
    --configuration-directory\n\
    --config-dir      - Load <dir>/*.conf as configuration files.\n\
    --default-configuration\n\
    --def-configuration\n\
                      - Do the default configuration, which is done if no\n\
                        options are given.\n\
\n\
    --configuration-data-daemon\n\
    --config-data-daemon\n\
                      - Parse configuration given in the\n\
                        org.and.daemon-conf-1.0 namespace.\n\
    --daemon          - Toggle becoming a daemon%s.\n\
    --chroot          - Change root.\n\
    --drop-privs      - Toggle droping privilages%s.\n\
    --priv-uid        - Drop privilages to this uid.\n\
    --priv-gid        - Drop privilages to this gid.\n\
    --pid-file        - Log pid to file.\n\
    --cntl-file       - Create control file.\n\
    --accept-filter-file\n\
                      - Load Linux Socket Filter code for accept().\n\
    --processes       - Number of processes to use (default: 1).\n\
    --debug -d        - Raise debug level (can be used upto 3 times).\n\
    --local           - Local filename to bind (default: \"all\").\n\
    --help -h         - Print this message.\n\
    --max-connections\n\
                   -M - Max connections allowed (0 = no limit).\n\
    --idle-timeout -t - Timeout (usecs) for connections that are idle.\n\
    --version -V      - Print the version string.\n\
\n\
  Dispatcher plugin audisp2file\n\
    --configuration-data-audisp2file\n\
    --config-data-audisp2file\n\
                      - Parse configuration given in the\n\
                        com.redhat.audisp2file-conf-1.0 namespace.\n\
    --input           - Socket to get audisp events from.\n\
",
               prefix, program_name,
               opt_def_toggle(FALSE), opt_def_toggle(FALSE));

  if (io_put_all(out, ret ? STDERR_FILENO : STDOUT_FILENO) == IO_FAIL)
    err(EXIT_FAILURE, "write");
  
  exit (ret);
}

/* HACK ... */
extern const char *audit_msg_type_to_name(int msg_type);

static int audisp__fmt_add_msg_type2name(Vstr_base *base, size_t pos,
                                         Vstr_fmt_spec *spec)
{
  unsigned int type = VSTR_FMT_CB_ARG_VAL(spec, unsigned int, 0);
  const char *cstr = NULL;

  if (!(cstr = audit_msg_type_to_name(type)))
    cstr = "Unknown";
  
  return (vstr_add_cstr_buf(base, pos, cstr));
}

static int audisp_fmt_add_msg_type2name(Vstr_conf *conf, const char *name)
{
  return (vstr_fmt_add(conf, name, audisp__fmt_add_msg_type2name,
                       VSTR_TYPE_FMT_UINT,
                       VSTR_TYPE_FMT_END));

}

static void serv_init(void)
{
  if (!vstr_init()) /* init the library */
    errno = ENOMEM, err(EXIT_FAILURE, "init");

  vlg_init();

  if (!vstr_cntl_conf(NULL, VSTR_CNTL_CONF_SET_TYPE_GRPALLOC_CACHE,
                      VSTR_TYPE_CNTL_CONF_GRPALLOC_CSTR))
    errno = ENOMEM, err(EXIT_FAILURE, "init");

  if (!vstr_cntl_conf(NULL,
                      VSTR_CNTL_CONF_SET_NUM_BUF_SZ, OPT_SERV_CONF_BUF_SZ))
    errno = ENOMEM, err(EXIT_FAILURE, "init");
  
  /* no passing of conf to evnt */
  if (!vstr_cntl_conf(NULL, VSTR_CNTL_CONF_SET_FMT_CHAR_ESC, '$') ||
      !vstr_cntl_conf(NULL, VSTR_CNTL_CONF_SET_LOC_CSTR_THOU_SEP, "_") ||
      !vstr_cntl_conf(NULL, VSTR_CNTL_CONF_SET_LOC_CSTR_THOU_GRP, "\3") ||
      !vstr_sc_fmt_add_all(NULL) ||
      !vlg_sc_fmt_add_all(NULL) ||
      !VSTR_SC_FMT_ADD(NULL, audisp_fmt_add_msg_type2name,
                       "<au-msg_type2name", "u", ">") /* ||
       !VSTR_SC_FMT_ADD(NULL, http_fmt_add_vstr_add_vstr,
                       "<http-esc.vstr", "p%zu%zu", ">") ||
      !VSTR_SC_FMT_ADD(NULL, http_fmt_add_vstr_add_sect_vstr,
      "<http-esc.vstr.sect", "p%p%u", ">") */ )
    errno = ENOMEM, err(EXIT_FAILURE, "init");

  if (!(vlg = vlg_make()))
    errno = ENOMEM, err(EXIT_FAILURE, "init");

  if (!VSTR_SC_FMT_ADD(vlg->out_vstr->conf, audisp_fmt_add_msg_type2name,
                       "<au-msg_type2name", "u", ">"))
    errno = ENOMEM, err(EXIT_FAILURE, "init");
  
  /*
  if (!VSTR_SC_FMT_ADD(vlg->out_vstr->conf, http_fmt_add_vstr_add_vstr,
                       "<http-esc.vstr", "p%zu%zu", ">") ||
      !VSTR_SC_FMT_ADD(vlg->out_vstr->conf, http_fmt_add_vstr_add_sect_vstr,
                       "<http-esc.vstr.sect", "p%p%u", ">"))
    errno = ENOMEM, err(EXIT_FAILURE, "init");
  if (!VSTR_SC_FMT_ADD(vlg->sig_out_vstr->conf, http_fmt_add_vstr_add_vstr,
                       "<http-esc.vstr", "p%zu%zu", ">") ||
      !VSTR_SC_FMT_ADD(vlg->sig_out_vstr->conf, http_fmt_add_vstr_add_sect_vstr,
                       "<http-esc.vstr.sect", "p%p%u", ">"))
    errno = ENOMEM, err(EXIT_FAILURE, "init");
  */
  
  if (!socket_poll_init(0, SOCKET_POLL_TYPE_MAP_DIRECT))
      errno = ENOMEM, err(EXIT_FAILURE, "init");
  
  evnt_logger(vlg);
  evnt_timeout_init();

  vlg_time_set(vlg, evnt_sc_time);
  
  opt_serv_logger(vlg);

  /* httpd_init(vlg); */
  
  opt_serv_sc_signals();
}

static struct Evnt *audisp2file__input = NULL;
static Vstr_base *audisp2file__output = NULL;
static int audisp2file__fd = -1;

struct Audisp_hdr 
{ /* copy of struct audit_dispatcher_header */
 uint32_t        vers;    /* The version of this protocol */
 uint32_t        hlen;   /* Header length */
 uint32_t        type;   /* Message type */
 uint32_t        size;   /* Size of data following the header */
};

static int audisp2file__cb_func_recv(struct Evnt *evnt)
{ /* main audit dispatcher socket ... input for everything */
  Vstr_base *data = evnt->io_r;
  unsigned int ern = 0;
  struct Audisp_hdr hdr[1];
  
  ASSERT(!evnt->io_r_shutdown);
  ASSERT(evnt == audisp2file__input);

  if (!evnt_recv(evnt, &ern))
  {
    if (ern != VSTR_TYPE_SC_READ_FD_ERR_EOF)
    {
      vlg_dbg2(vlg, "RECV ERR from[$<sa:%p>]: %u\n", EVNT_SA(evnt), ern);
      goto con_cleanup;
    }
    if (!evnt_shutdown_r(evnt, TRUE))
      goto con_cleanup;

    return (FALSE);
  }

  while (data->len >= sizeof(struct Audisp_hdr))
  {
    vstr_export_buf(data, 1, sizeof(struct Audisp_hdr), hdr, sizeof(hdr));
    
    if ((hdr->vers != 0) || (hdr->hlen != sizeof(struct Audisp_hdr)))
    {
      vlg_warn(vlg,
               "RECV ERR from[$<sa:%p>] vers[%u] hsz[%u] msg-type[%u:$<au-msg_type2name:%u>] sz[%u]\n",
             EVNT_SA(evnt),
             (unsigned int)hdr->vers,
             (unsigned int)hdr->hlen,
             (unsigned int)hdr->type,
             (unsigned int)hdr->type,
             (unsigned int)hdr->size);

      return (FALSE);
    }

    if (hdr->size > (data->len - hdr->hlen))
      break; /* not enough data, yet */

    /* FIXME: print output in nice format to different file types */
    evnt_got_pkt(evnt);

    vstr_add_fmt(audisp2file__output, audisp2file__output->len,
                 "$<au-msg_type2name:%u>(%u) msg=%u:${vstr:%p%zu%zu%u},\n",
                 /* evnt->acct.req_got, */
                 (unsigned int)hdr->type,
                 (unsigned int)hdr->type,
                 (unsigned int)hdr->size,
                 data, (size_t)(hdr->hlen + 1), (size_t)hdr->size,
                 (unsigned int)VSTR_TYPE_ADD_DEF);
  
    /* FIXME: move to copying ptrs only ... not all data */
    vstr_del(data, 1, hdr->hlen + hdr->size);
  }

  if (data->conf->malloc_bad)
    goto con_cleanup;
  if (audisp2file__output->conf->malloc_bad)
    goto con_cleanup;

  if (io_put_all(audisp2file__output, audisp2file__fd) == IO_FAIL)
    goto con_cleanup;
  
  return (TRUE);
  
 con_cleanup:
  evnt->io_r->conf->malloc_bad = FALSE;
  evnt->io_w->conf->malloc_bad = FALSE;
    
  return (FALSE);
}

static void audisp2file__cb_func_free(struct Evnt *evnt)
{ /* main audit dispatcher socket ... input for everything, once this is gone
     give up */
  ASSERT(evnt == audisp2file__input);

  ASSERT(audisp2file__fd != -1);
  ASSERT(audisp2file__output);
  
  evnt_vlg_stats_info(evnt, "MAIN FREE");

  vstr_free_base(audisp2file__output);
  if (audisp2file__fd != STDOUT_FILENO)
    close(audisp2file__fd);

  audisp2file__fd     = -1;
  audisp2file__output = NULL;
  audisp2file__input  = NULL;

  F(evnt);
}

static int audisp2file__cb_func_connect(struct Evnt *evnt)
{
  ASSERT(evnt == audisp2file__input);
  
  evnt_pkt_wait_r(evnt);
  
  return (TRUE);
}

static const struct Evnt_cbs audisp2file__main_evnt_cbs =
{
 evnt_cb_func_accept,    /* def */
 audisp2file__cb_func_connect,

 audisp2file__cb_func_recv,
 evnt_cb_func_send,      /* def */

 audisp2file__cb_func_free,
 evnt_cb_func_shutdown_r /* def */
};

static void serv_make_main_input(const char *program_name)
{
  int fd = -1;
  
  if (!(audisp2file__input = MK(sizeof(struct Evnt))))
    errno = ENOMEM, err(EXIT_FAILURE, "make stdin Evnt");

  if (!audisp2file_opts->main_input->len)
  { /* use stdin */
    if ((fd = dup(STDIN_FILENO)) == -1)
      err(EXIT_FAILURE, "dup stdin");

    if (!evnt_make_custom(audisp2file__input, fd, NULL, POLLIN))
      errx(EXIT_FAILURE, "make custom stdin");
  }
  else
  {
    const char *main_input = NULL;

    OPT_SC_EXPORT_CSTR(main_input, audisp2file_opts->main_input, FALSE,
                       "main input socket");
    
    if (!evnt_make_con_local(audisp2file__input, main_input))
      err(EXIT_FAILURE, "make local %s", main_input);
    audisp2file__input->flag_q_pkt_move = FALSE;
    
    if (audisp2file__input->flag_q_none)
      audisp2file__cb_func_connect(audisp2file__input);
  }
  
  close(STDIN_FILENO);
  if (open("/dev/null", O_RDONLY) != STDIN_FILENO)
    err(EXIT_FAILURE, "open stdin");
  
  audisp2file__input->cbs = &audisp2file__main_evnt_cbs;
}


static int serv_def_conf(Vstr_base *out)
{ /* pretend we passed (if it works...):
   * -C /etc/audit/audisp2file.conf --conf-dir /etc/audit/audisp2file-conf.d ... */
  struct stat64 d_stat[1];
  static const char pc_file[] = PATH_SYSCONFDIR "/audit/audispd-plugins.d/audisp2file.conf";
  static const char pc_dir[]  = PATH_SYSCONFDIR "/audit/audispd-plugins.d/audisp2file-conf.d";
  Opt_serv_opts *opts = audisp2file_opts->s;

  if ((stat64(pc_file, d_stat) != -1) &&
      !audisp2file_conf_parse_file(out, audisp2file_opts, pc_file))
    return (FALSE);
  if ((stat64(pc_dir,  d_stat) != -1) &&
      !opt_serv_sc_config_dir(out, opts, pc_dir,
                              audisp2file_sc_conf_parse_dir_file))
    return (FALSE);
  
  return (TRUE);
}

#define POPT_NUM_NR_ARG(x, y) do {                                      \
      Opt_serv_policy_opts *popt_scan = audisp2file_opts->s->def_policy;    \
                                                                        \
      while (popt_scan)                                                 \
      {                                                                 \
        Audisp2file_policy_opts *popt_tmp = (Audisp2file_policy_opts *)popt_scan;   \
        OPT_NUM_NR_ARG(popt_tmp->x, y);                                 \
        popt_scan = popt_scan->next;                                    \
      }                                                                 \
    } while (FALSE)


static void serv_cmd_line(int argc, char *argv[])
{
  int optchar = 0;
  const char *program_name = NULL;
  struct option long_options[] =
  {
   OPT_SERV_DECL_GETOPTS(),
   
   {"configuration-file", required_argument, NULL, 'C'},
   {"config-file",        required_argument, NULL, 'C'},
   {"conf-file",          required_argument, NULL, 'C'},
   {"configuration-directory", required_argument, NULL, 140},
   {"config-dir",              required_argument, NULL, 140},
   {"conf-dir",                required_argument, NULL, 140},
   {"configuration-data-daemon", required_argument, NULL, 143},
   {"config-data-daemon",        required_argument, NULL, 143},
   {"configuration-data-audisp2file", required_argument, NULL, 144},
   {"config-data-audisp2file",        required_argument, NULL, 144},
   {"default-configuration", no_argument, NULL, 149},
   {"def-config",            no_argument, NULL, 149},
   
   {"max-header-sz", required_argument, NULL, 128},
   {"main-input", required_argument, NULL, 129},
   {"input", required_argument, NULL, 129},
   /* 130 */
   /* 131 */
   /* 132 */
   /* 133 */
   /* 134 */
   /* 135 */
   /* 136 */
   /* 137 */
   /* 138 */
   /* 139 */
   /* 140 -- config. dir above */
   /* 141 */
   /* 142 */
   /* 143 -- config data above */
   /* 144 -- config data above */
   /* 145 */
   /* 146 */
   /* 147 */
   /* 148 */
   /* 149 -- config data above */
   {NULL, 0, NULL, 0}
  };
  const char *chroot_dir = NULL;
  Vstr_base *out = vstr_make_base(NULL);
  Opt_serv_opts *opts = audisp2file_opts->s;
  Audisp2file_policy_opts *popts = NULL;
  
  if (!out)
    errno = ENOMEM, err(EXIT_FAILURE, "command line");

  /* evnt_opt_nagle = TRUE; */
  
  program_name = opt_program_name(argv[0], PACKAGE);
  opts->name_cstr = program_name;
  opts->name_len  = strlen(program_name);

  if (!audisp2file_conf_init(audisp2file_opts))
    errno = ENOMEM, err(EXIT_FAILURE, "options");
  
  if (!geteuid()) /* If root */
    vstr_add_cstr_ptr(opts->addr_beg->acpt_address, 0, AUDISP2FILE_CONF_DEF_LISTEN);
  else /* uncommon unprived default */
    vstr_add_cstr_ptr(opts->addr_beg->acpt_address, 0,
                      AUDISP2FILE_CONF_DEF_USR_LISTEN);
    
  popts = (Audisp2file_policy_opts *)opts->def_policy;
  while ((optchar = getopt_long(argc, argv, "C:dhH:M:nP:t:V",
                                long_options, NULL)) != -1)
  {
    switch (optchar)
    {
      case '?': usage(program_name, EXIT_FAILURE, "");
      case 'h': usage(program_name, EXIT_SUCCESS, "");
        
      case 'V':
        vstr_add_fmt(out, 0, " %s version %s.\n", program_name, VERSION);
        
        if (io_put_all(out, STDOUT_FILENO) == IO_FAIL)
          err(EXIT_FAILURE, "write");
        
        exit (EXIT_SUCCESS);
      
      OPT_SERV_GETOPTS(opts);

      case 'C':
        if (!audisp2file_conf_parse_file(out, audisp2file_opts, optarg))
          goto out_err_conf_msg;
        break;
      case 140:
        if (!opt_serv_sc_config_dir(out, opts, optarg,
                                    audisp2file_sc_conf_parse_dir_file))
          goto out_err_conf_msg;
        break;
      case 143:
        if (!opt_serv_conf_parse_cstr(out, opts, optarg))
          goto out_err_conf_msg;
      break;
      case 144:
        if (!audisp2file_conf_parse_cstr(out, audisp2file_opts, optarg))
          goto out_err_conf_msg;
        break;
      case 149:
        if (!serv_def_conf(out))
          goto out_err_conf_msg;
        break;
        
      case 128: POPT_NUM_NR_ARG(max_header_sz, "max header size"); break;
      case 129: OPT_VSTR_ARG(audisp2file_opts->main_input);
      
      ASSERT_NO_SWITCH_DEF();
    }
  }

  ASSERT(optind >= 1);
  if (optind <= 1) /* just program name, no options so use default... */
    if (!serv_def_conf(out))
      goto out_err_conf_msg;
  vstr_free_base(out); out = NULL;
  
  argc -= optind;
  argv += optind;

  if (argc > 1)
    usage(program_name, EXIT_FAILURE, " Too many arguments.\n");

  if (!(audisp2file__output = vstr_make_base(NULL)))
    errno = ENOMEM, err(EXIT_FAILURE, "init");

  if (argc == 0)
    audisp2file__fd = STDOUT_FILENO;
  else if (!strcmp(argv[0], "-"))
    audisp2file__fd = STDOUT_FILENO;
  else
  { /* mode/create should be configurable... */
    audisp2file__fd = open(argv[0], O_WRONLY | O_CREAT | O_APPEND, 0600);
    if (audisp2file__fd == -1)
      err(EXIT_FAILURE,
          "open(%s, O_WRONLY | O_CREAT | O_APPEND, 0600)", argv[0]);
  }

  if (!audisp2file_opts->main_input->len && (audisp2file__fd == STDOUT_FILENO))
    usage(program_name, EXIT_FAILURE, " Can't use STDIN and STDOUT at once.\n");
  
  {
    const char *pid_file = NULL;
    
    /* FIXME: move somewhere else */
    if (!opts->poll_backend->len)
      evnt_poll = &EVNT_POLL_BE_DEF;
    else if (VALLEQ(opts->poll_backend, "poll"))
      evnt_poll = &evnt_poll_be_poll;
    else if (VALLEQ(opts->poll_backend, "epoll"))
      evnt_poll = &evnt_poll_be_epoll;
    else
      usage(program_name, EXIT_FAILURE,
            " Not specified a valid poll backend.\n");
    evnt_poll->init();
  
    vlg_syslog_facility_set(vlg, opts->syslog_facility);
    /* vlg_date_set(vlg, opts->vlg_date_fmt_type); */
    if (!opts->vlg_tweaked_size)
    {
      if (opts->become_daemon)      
        opts->vlg_size = OPT_SERV_CONF_VLG_SIZE_UNTWEAKED_DAEMON;
      else
        opts->vlg_size = OPT_SERV_CONF_VLG_SIZE_UNTWEAKED_CONSOLE;
    }
    vlg_size_set(vlg, opts->vlg_size);
    vlg_syslog_native_set(vlg, opts->vlg_syslog_native);
  
    OPT_SC_EXPORT_CSTR(pid_file,   opts->pid_file,   FALSE, "pid file");
    OPT_SC_EXPORT_CSTR(chroot_dir, opts->chroot_dir, FALSE, "chroot directory");
    
    if (opts->drop_privs)
    {
      OPT_SERV_SC_RESOLVE_UID(opts);
      OPT_SERV_SC_RESOLVE_GID(opts);
    }
  
    serv_make_main_input(program_name);
    
    if (opts->become_daemon)
    {
      /* FIXME: should be better ... more config. options
         
        if (daemon(FALSE, FALSE) == -1)
        err(EXIT_FAILURE, "daemon");
      */
      
      chdir("/"); /* don't fork, or setsid */

      dup2(STDIN_FILENO, STDOUT_FILENO); /* STDIN was done in make_main_input */
      dup2(STDIN_FILENO, STDERR_FILENO);
      
      vlg_daemon(vlg, program_name);
    }
    
    /* NOTE: after daemon so don't use err() anymore ... */
    
    if (opts->rlim_as_call)
      opt_serv_sc_rlim_as_num(opts->rlim_as_num);
    if (opts->rlim_core_call)
      opt_serv_sc_rlim_core_num(opts->rlim_core_num);
    if (opts->rlim_file_call)
      opt_serv_sc_rlim_file_num(opts->rlim_file_num);
  
    if (pid_file)
      vlg_pid_file(vlg, pid_file);

    if (opts->cntl_file->len)
      cntl_make_file(vlg, opts->cntl_file);

    if (chroot_dir)
    { /* preload locale info. so syslog can log in localtime, this doesn't work
       * with glibc's interfaces ...
       * but we have our own and gmtime_r() / localime_r() etc. */
      time_t now = time(NULL);
      (void)localtime(&now);
      
      vlg_sc_bind_mount(chroot_dir);
    }
    
    if (chroot_dir && ((chroot(chroot_dir) == -1) || (chdir("/") == -1)))
      vlg_err(vlg, EXIT_FAILURE, "chroot(%s): %m\n", chroot_dir);
    
    if (opts->drop_privs)
    {
      if (opts->keep_cap_fowner && (PROC_CNTL_KEEPCAPS(TRUE) == -1))
      {
        vlg_warn(vlg, "prctl(%s, %s): %m\n", "PR_SET_KEEPCAPS", "TRUE");
        opts->keep_cap_fowner = FALSE;
      }
      
      opt_serv_sc_drop_privs(opts);
      
      if (opts->keep_cap_fowner)
      { /* use + instead of = because that's how getpcaps prints it,
         * however they do the same due to no input */
        cap_t caps = cap_from_text("cap_fowner+ep-i");

        if (!caps)
          vlg_err(vlg, EXIT_FAILURE, "cap_from_text(%s): %m\n",
                  "cap_fowner+pe-i");
        
        if (PROC_CNTL_KEEPCAPS(FALSE) == -1)
          vlg_err(vlg, EXIT_FAILURE,
                  "prctl(%s, %s): %m\n", "PR_SET_KEEPCAPS", "FALSE");
        
        if (cap_set_proc(caps) == -1)
          vlg_err(vlg, EXIT_FAILURE, "cap_set_proc(%s): %m\n",
                  "cap_fowner+ep-i");
        if (cap_free(caps) == -1)
          vlg_err(vlg, EXIT_FAILURE, "cap_free(): %m\n");
      }
    }

    if (FALSE && (opts->num_procs > 1))
    {
      int has_cntl = !!opts->cntl_file->len;
      cntl_sc_multiproc(vlg, opts->num_procs, has_cntl, opts->use_pdeathsig);
    }
    
    if (opts->make_dumpable && (PROC_CNTL_DUMPABLE(TRUE) == -1))
      vlg_warn(vlg, "prctl(%s, %s): %m\n", "PR_SET_DUMPABLE", "TRUE");
  }

  audisp2file_opts->beg_time = time(NULL);
  
  {
    struct Evnt *evnt = evnt_queue("recv");
    
    while (evnt)
    {
      vlg_info(vlg, "READY [$<sa:%p>]: %s\n", EVNT_SA(evnt),
               chroot_dir ? chroot_dir : "");
      evnt = evnt->next;
    }
  }

  opt_serv_conf_free_beg(opts);
  return;

 out_err_conf_msg:
  vstr_add_cstr_ptr(out, out->len, "\n");
  if (io_put_all(out, STDERR_FILENO) == IO_FAIL)
    err(EXIT_FAILURE, "write");
  exit (EXIT_FAILURE);  
}

#if COMPILE_DEBUG
static int cov(void) /* for coverage */
{
  Opt_serv_policy_opts *opt = NULL;
  
  assert((F(M0(1, 1)), 1));
  
  ASSERT(date_syslog_trad(vlg->dt, time(NULL)));

  opt = opt_policy_make(audisp2file_opts->s);
  vstr_ref_del(opt->ref);
  
  return (TRUE);
}
#endif

int main(int argc, char *argv[])
{
  if (sizeof(uintmax_t) != sizeof(VSTR_AUTOCONF_uintmax_t))
    errx(EXIT_FAILURE, "uintmax_t size is different between program and Vstr");

  serv_init();

  serv_cmd_line(argc, argv);

  ASSERT(cov()); /* for coverage */
  
  while (evnt_waiting())
  {
    evnt_sc_main_loop(AUDISP2FILE_CONF_MAX_WAIT_SEND);
    opt_serv_sc_check_children();
    opt_serv_sc_cntl_resources(audisp2file_opts->s);
  }
  evnt_out_dbg3("E");
  
  evnt_timeout_exit();
  cntl_child_free();
  evnt_close_all();
  
  vlg_free(vlg);
  vlg_exit();

  audisp2file_conf_free(audisp2file_opts);
  
  timer_q_exit();
  
  vstr_exit();

  MALLOC_CHECK_EMPTY();
  
  exit (EXIT_SUCCESS);
}
